package dev.pfaff.lazy;

import java.lang.invoke.*;
import java.util.function.Supplier;

import static java.lang.invoke.MethodHandles.constant;
import static java.lang.invoke.MethodHandles.insertArguments;
import static java.lang.invoke.MethodType.methodType;

public final class Lazy<T> implements Supplier<T> {
	private static final MethodHandle MH_lazyInit;
	private static final MethodHandle MH_lazyInitSynchronized;

	static {
		try {
			MH_lazyInit = MethodHandles.lookup().findStatic(Lazy.class,
					"lazyInit",
					methodType(Object.class, CallSite.class, Supplier.class));
			MH_lazyInitSynchronized = MethodHandles.lookup().findStatic(Lazy.class,
					"lazyInitSynchronized",
					methodType(Object.class, CallSite.class, InitializationStatus.class, Supplier.class));
		} catch (ReflectiveOperationException e) {
			throw new RuntimeException(e);
		}
	}

	private static <T> T get(MethodHandle getter) {
		try {
			//noinspection unchecked
			return (T) getter.invoke();
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	private final MethodHandle getter;

	private static <T> T lazyInit(CallSite callSite, Supplier<? extends T> supplier) {
		var value = supplier.get();
		callSite.setTarget(constant(callSite.type().returnType(), value));
		return value;
	}

	private static <T> T lazyInitSynchronized(CallSite callSite, InitializationStatus status, Supplier<? extends T> supplier) {
		synchronized (callSite) {
			if (status.initialized) return get(callSite.getTarget());
			var value = lazyInit(callSite, supplier);
			status.initialized = true;
			return value;
		}
	}

	public static <T> MethodHandle makeLazyConstantHandle(Class<T> type, Supplier<? extends T> supplier, boolean isSynchronized) {
		var callSite = new MutableCallSite(isSynchronized
				? insertArguments(
				MH_lazyInitSynchronized.asType(MH_lazyInitSynchronized.type().changeReturnType(type)),
				1, new InitializationStatus(), supplier)
				: insertArguments(MH_lazyInit.asType(MH_lazyInit.type().changeReturnType(type)), 1, supplier));
		return callSite.dynamicInvoker();
	}

	public Lazy(Supplier<? extends T> supplier, boolean isSynchronized) {
		this.getter = makeLazyConstantHandle(Object.class, supplier, isSynchronized);
	}

	public Lazy(Supplier<T> supplier) {
		this(supplier, false);
	}

	@Override
	public T get() {
		return get(getter);
	}

	private static final class InitializationStatus {
		private boolean initialized;
	}
}
